@@ -1,4 +1,5 @@
 use crate::call::{AsyncCaller, SyncCaller};
+use candid::utils::ArgumentEncoder;
 use candid::{parser::value::IDLValue, ser::IDLBuilder, utils::ArgumentDecoder, CandidType};
 use garcon::Waiter;
 use ic_agent::{ic_types::Principal, Agent, AgentError, RequestId};
@@ -23,14 +24,18 @@ pub enum CanisterBuilderError {
 }
 
 /// A canister builder, which can be used to create a canister abstraction.
-#[derive(Debug)]
-pub struct CanisterBuilder<'agent, T = ()> {
+#[derive(Debug, Default)]
+pub struct CanisterBuilder<'agent> {
     agent: Option<&'agent Agent>,
     canister_id: Option<Result<Principal, CanisterBuilderError>>,
-    interface: T,
 }
 
-impl<'agent, T> CanisterBuilder<'agent, T> {
+impl<'agent> CanisterBuilder<'agent> {
+    /// Create a canister builder with no value.
+    pub fn new() -> CanisterBuilder<'static> {
+        Default::default()
+    }
+
     /// Attach a canister ID to this canister.
     pub fn with_canister_id<E, P>(self, canister_id: P) -> Self
     where
@@ -56,7 +61,7 @@ impl<'agent, T> CanisterBuilder<'agent, T> {
     }
 
     /// Create this canister abstraction after passing in all the necessary state.
-    pub fn build(self) -> Result<Canister<'agent, T>, CanisterBuilderError> {
+    pub fn build(self) -> Result<Canister<'agent>, CanisterBuilderError> {
         let canister_id = if let Some(cid) = self.canister_id {
             cid?
         } else {
@@ -66,38 +71,7 @@ impl<'agent, T> CanisterBuilder<'agent, T> {
         let agent = self
             .agent
             .ok_or(CanisterBuilderError::MustSpecifyAnAgent())?;
-        Ok(Canister {
-            agent,
-            canister_id,
-            interface: self.interface,
-        })
-    }
-}
-
-impl Default for CanisterBuilder<'static, ()> {
-    fn default() -> Self {
-        CanisterBuilder {
-            agent: None,
-            canister_id: None,
-            interface: (),
-        }
-    }
-}
-
-impl<'agent> CanisterBuilder<'agent, ()> {
-    /// Create a canister builder with no value.
-    pub fn new() -> CanisterBuilder<'static, ()> {
-        Default::default()
-    }
-
-    /// Apply an interface to this canister. An interface can add methods to the canister's
-    /// type. For example, see the Management Canister.
-    pub fn with_interface<T>(self, interface: T) -> CanisterBuilder<'agent, T> {
-        CanisterBuilder {
-            agent: self.agent,
-            canister_id: self.canister_id,
-            interface,
-        }
+        Ok(Canister { agent, canister_id })
     }
 }
 
@@ -107,30 +81,23 @@ impl<'agent> CanisterBuilder<'agent, ()> {
 ///
 /// This is the higher level construct for talking to a canister on the Internet
 /// Computer.
-#[derive(Debug)]
-pub struct Canister<'agent, T = ()> {
+#[derive(Debug, Clone)]
+pub struct Canister<'agent> {
     pub(super) agent: &'agent Agent,
     pub(super) canister_id: Principal,
-    interface: T,
 }
 
-impl<'agent, T> Canister<'agent, T> {
+impl<'agent> Canister<'agent> {
     /// Get the canister ID of this canister.
     pub fn canister_id_<'canister: 'agent>(&'canister self) -> &Principal {
         &self.canister_id
     }
 
-    /// Get the interface object from this canister. Sometimes those interfaces might have
-    /// custom methods that are useful.
-    pub fn interface_<'canister: 'agent>(&'canister self) -> &T {
-        &self.interface
-    }
-
     /// Create an AsyncCallBuilder to do an update call.
     pub fn update_<'canister: 'agent>(
         &'canister self,
         method_name: &str,
-    ) -> AsyncCallBuilder<'agent, 'canister, T> {
+    ) -> AsyncCallBuilder<'agent, 'canister> {
         AsyncCallBuilder::new(self, method_name)
     }
 
@@ -138,7 +105,7 @@ impl<'agent, T> Canister<'agent, T> {
     pub fn query_<'canister: 'agent>(
         &'canister self,
         method_name: &str,
-    ) -> SyncCallBuilder<'agent, 'canister, T> {
+    ) -> SyncCallBuilder<'agent, 'canister> {
         SyncCallBuilder::new(self, method_name)
     }
 
@@ -156,25 +123,17 @@ impl<'agent, T> Canister<'agent, T> {
             .wait(request_id, self.canister_id, disable_range_check, waiter)
             .await
     }
-}
 
-impl<'agent, T> Canister<'agent, T>
-where
-    T: Clone,
-{
     /// Creates a copy of this canister, changing the canister ID to the provided principal.
     pub fn clone_with_(&self, id: Principal) -> Self {
         Self {
             agent: self.agent,
             canister_id: id,
-            interface: self.interface.clone(),
         }
     }
-}
 
-impl<'agent> Canister<'agent, ()> {
     /// Create a CanisterBuilder instance to build a canister abstraction.
-    pub fn builder() -> CanisterBuilder<'agent, ()> {
+    pub fn builder() -> CanisterBuilder<'agent> {
         Default::default()
     }
 }
@@ -199,7 +158,7 @@ impl fmt::Debug for ArgumentType {
     }
 }
 
-/// A builder for a canister argument, allowing you to append elements to [`ArgumentType`] with chaining syntax.
+/// A builder for a canister argument, allowing you to append elements to an argument tuple with chaining syntax.
 #[derive(Debug)]
 pub struct Argument(Result<ArgumentType, AgentError>);
 
@@ -265,11 +224,32 @@ impl Argument {
     pub fn reset(&mut self) {
         *self = Default::default();
     }
+
+    /// Creates an empty argument.
+    pub fn new() -> Self {
+        Default::default()
+    }
+
+    /// Creates an argument from an arbitrary blob. Equivalent to [`set_raw_arg`].
+    pub fn from_raw(raw: Vec<u8>) -> Self {
+        Self(Ok(ArgumentType::Raw(raw)))
+    }
+
+    /// Creates an argument from an existing Candid ArgumentEncoder.
+    pub fn from_candid(tuple: impl ArgumentEncoder) -> Self {
+        let mut builder = IDLBuilder::new();
+        Self(
+            tuple
+                .encode(&mut builder)
+                .map(|_| ArgumentType::Idl(builder))
+                .map_err(|e| AgentError::CandidError(Box::new(e))),
+        )
+    }
 }
 
 impl Default for Argument {
     fn default() -> Self {
-        Argument(Ok(ArgumentType::Idl(IDLBuilder::new())))
+        Self(Ok(ArgumentType::Idl(IDLBuilder::new())))
     }
 }
 
@@ -277,17 +257,17 @@ impl Default for Argument {
 ///
 /// See [SyncCaller] for a description of this structure once built.
 #[derive(Debug)]
-pub struct SyncCallBuilder<'agent, 'canister: 'agent, T> {
-    canister: &'canister Canister<'agent, T>,
+pub struct SyncCallBuilder<'agent, 'canister: 'agent> {
+    canister: &'canister Canister<'agent>,
     method_name: String,
     effective_canister_id: Principal,
     arg: Argument,
 }
 
-impl<'agent, 'canister: 'agent, T> SyncCallBuilder<'agent, 'canister, T> {
+impl<'agent, 'canister: 'agent> SyncCallBuilder<'agent, 'canister> {
     /// Create a new instance of an AsyncCallBuilder.
     pub(super) fn new<M: Into<String>>(
-        canister: &'canister Canister<'agent, T>,
+        canister: &'canister Canister<'agent>,
         method_name: M,
     ) -> Self {
         Self {
@@ -299,13 +279,10 @@ impl<'agent, 'canister: 'agent, T> SyncCallBuilder<'agent, 'canister, T> {
     }
 }
 
-impl<'agent, 'canister: 'agent, Interface> SyncCallBuilder<'agent, 'canister, Interface> {
+impl<'agent, 'canister: 'agent> SyncCallBuilder<'agent, 'canister> {
     /// Add an argument to the candid argument list. This requires Candid arguments, if
     /// there is a raw argument set (using [with_arg_raw]), this will fail.
-    pub fn with_arg<Argument>(
-        mut self,
-        arg: Argument,
-    ) -> SyncCallBuilder<'agent, 'canister, Interface>
+    pub fn with_arg<Argument>(mut self, arg: Argument) -> SyncCallBuilder<'agent, 'canister>
     where
         Argument: CandidType + Sync + Send,
     {
@@ -316,17 +293,14 @@ impl<'agent, 'canister: 'agent, Interface> SyncCallBuilder<'agent, 'canister, In
     /// Add an argument to the candid argument list. This requires Candid arguments, if
     /// there is a raw argument set (using [with_arg_raw]), this will fail.
     /// TODO: make this method unnecessary https://github.com/dfinity/agent-rs/issues/132
-    pub fn with_value_arg(
-        mut self,
-        arg: IDLValue,
-    ) -> SyncCallBuilder<'agent, 'canister, Interface> {
+    pub fn with_value_arg(mut self, arg: IDLValue) -> SyncCallBuilder<'agent, 'canister> {
         self.arg.push_value_arg(arg);
         self
     }
 
     /// Replace the argument with raw argument bytes. This will overwrite the current
     /// argument set, so calling this method twice will discard the first argument.
-    pub fn with_arg_raw(mut self, arg: Vec<u8>) -> SyncCallBuilder<'agent, 'canister, Interface> {
+    pub fn with_arg_raw(mut self, arg: Vec<u8>) -> SyncCallBuilder<'agent, 'canister> {
         self.arg.set_raw_arg(arg);
         self
     }
@@ -335,7 +309,7 @@ impl<'agent, 'canister: 'agent, Interface> SyncCallBuilder<'agent, 'canister, In
     pub fn with_effective_canister_id(
         mut self,
         canister_id: Principal,
-    ) -> SyncCallBuilder<'agent, 'canister, Interface> {
+    ) -> SyncCallBuilder<'agent, 'canister> {
         self.effective_canister_id = canister_id;
         self
     }
@@ -362,19 +336,19 @@ impl<'agent, 'canister: 'agent, Interface> SyncCallBuilder<'agent, 'canister, In
 ///
 /// See [AsyncCaller] for a description of this structure.
 #[derive(Debug)]
-pub struct AsyncCallBuilder<'agent, 'canister: 'agent, T> {
-    canister: &'canister Canister<'agent, T>,
+pub struct AsyncCallBuilder<'agent, 'canister: 'agent> {
+    canister: &'canister Canister<'agent>,
     method_name: String,
     effective_canister_id: Principal,
     arg: Argument,
 }
 
-impl<'agent, 'canister: 'agent, T> AsyncCallBuilder<'agent, 'canister, T> {
+impl<'agent, 'canister: 'agent> AsyncCallBuilder<'agent, 'canister> {
     /// Create a new instance of an AsyncCallBuilder.
     pub(super) fn new(
-        canister: &'canister Canister<'agent, T>,
+        canister: &'canister Canister<'agent>,
         method_name: &str,
-    ) -> AsyncCallBuilder<'agent, 'canister, T> {
+    ) -> AsyncCallBuilder<'agent, 'canister> {
         Self {
             canister,
             method_name: method_name.to_string(),
@@ -384,13 +358,10 @@ impl<'agent, 'canister: 'agent, T> AsyncCallBuilder<'agent, 'canister, T> {
     }
 }
 
-impl<'agent, 'canister: 'agent, Interface> AsyncCallBuilder<'agent, 'canister, Interface> {
+impl<'agent, 'canister: 'agent> AsyncCallBuilder<'agent, 'canister> {
     /// Add an argument to the candid argument list. This requires Candid arguments, if
     /// there is a raw argument set (using [with_arg_raw]), this will fail.
-    pub fn with_arg<Argument>(
-        mut self,
-        arg: Argument,
-    ) -> AsyncCallBuilder<'agent, 'canister, Interface>
+    pub fn with_arg<Argument>(mut self, arg: Argument) -> AsyncCallBuilder<'agent, 'canister>
     where
         Argument: CandidType + Sync + Send,
     {
@@ -400,7 +371,7 @@ impl<'agent, 'canister: 'agent, Interface> AsyncCallBuilder<'agent, 'canister, I
 
     /// Replace the argument with raw argument bytes. This will overwrite the current
     /// argument set, so calling this method twice will discard the first argument.
-    pub fn with_arg_raw(mut self, arg: Vec<u8>) -> AsyncCallBuilder<'agent, 'canister, Interface> {
+    pub fn with_arg_raw(mut self, arg: Vec<u8>) -> AsyncCallBuilder<'agent, 'canister> {
         self.arg.set_raw_arg(arg);
         self
     }
@@ -409,7 +380,7 @@ impl<'agent, 'canister: 'agent, Interface> AsyncCallBuilder<'agent, 'canister, I
     pub fn with_effective_canister_id(
         mut self,
         canister_id: Principal,
-    ) -> AsyncCallBuilder<'agent, 'canister, Interface> {
+    ) -> AsyncCallBuilder<'agent, 'canister> {
         self.effective_canister_id = canister_id;
         self
     }
@@ -461,12 +432,13 @@ mod tests {
             .unwrap();
         agent.fetch_root_key().await.unwrap();
 
-        let management_canister = Canister::builder()
-            .with_agent(&agent)
-            .with_canister_id("aaaaa-aa")
-            .with_interface(ManagementCanister)
-            .build()
-            .unwrap();
+        let management_canister = ManagementCanister::from_canister(
+            Canister::builder()
+                .with_agent(&agent)
+                .with_canister_id("aaaaa-aa")
+                .build()
+                .unwrap(),
+        );
 
         let (new_canister_id,) = management_canister
             .create_canister()
